@@ -126,13 +126,15 @@ size_t ngtcp2_crypto_packet_protection_ivlen(const ngtcp2_crypto_aead *aead) {
 }
 
 int ngtcp2_crypto_derive_packet_protection_key(
-    uint8_t *key, uint8_t *iv, uint8_t *hp_key, const ngtcp2_crypto_aead *aead,
-    const ngtcp2_crypto_md *md, const uint8_t *secret, size_t secretlen) {
+    uint8_t *key, uint8_t *iv, uint8_t *hp_key, uint32_t version,
+    const ngtcp2_crypto_aead *aead, const ngtcp2_crypto_md *md,
+    const uint8_t *secret, size_t secretlen) {
   static const uint8_t KEY_LABEL[] = "quic key";
   static const uint8_t IV_LABEL[] = "quic iv";
   static const uint8_t HP_KEY_LABEL[] = "quic hp";
   size_t keylen = ngtcp2_crypto_aead_keylen(aead);
   size_t ivlen = ngtcp2_crypto_packet_protection_ivlen(aead);
+  (void)version;
 
   if (ngtcp2_crypto_hkdf_expand_label(key, keylen, md, secret, secretlen,
                                       KEY_LABEL, sizeof(KEY_LABEL) - 1) != 0) {
@@ -217,8 +219,9 @@ int ngtcp2_crypto_derive_and_install_rx_key(ngtcp2_conn *conn, uint8_t *key,
   hp = &ctx->hp;
   ivlen = ngtcp2_crypto_packet_protection_ivlen(aead);
 
-  if (ngtcp2_crypto_derive_packet_protection_key(key, iv, hp_key, aead, md,
-                                                 secret, secretlen) != 0) {
+  if (ngtcp2_crypto_derive_packet_protection_key(
+          key, iv, hp_key, ngtcp2_conn_get_negotiated_version(conn), aead, md,
+          secret, secretlen) != 0) {
     return -1;
   }
 
@@ -351,8 +354,9 @@ int ngtcp2_crypto_derive_and_install_tx_key(ngtcp2_conn *conn, uint8_t *key,
   hp = &ctx->hp;
   ivlen = ngtcp2_crypto_packet_protection_ivlen(aead);
 
-  if (ngtcp2_crypto_derive_packet_protection_key(key, iv, hp_key, aead, md,
-                                                 secret, secretlen) != 0) {
+  if (ngtcp2_crypto_derive_packet_protection_key(
+          key, iv, hp_key, ngtcp2_conn_get_negotiated_version(conn), aead, md,
+          secret, secretlen) != 0) {
     return -1;
   }
 
@@ -415,7 +419,7 @@ int ngtcp2_crypto_derive_and_install_initial_key(
     ngtcp2_conn *conn, uint8_t *rx_secret, uint8_t *tx_secret,
     uint8_t *initial_secret, uint8_t *rx_key, uint8_t *rx_iv,
     uint8_t *rx_hp_key, uint8_t *tx_key, uint8_t *tx_iv, uint8_t *tx_hp_key,
-    const ngtcp2_cid *client_dcid) {
+    uint32_t version, const ngtcp2_cid *client_dcid) {
   uint8_t rx_secretbuf[NGTCP2_CRYPTO_INITIAL_SECRETLEN];
   uint8_t tx_secretbuf[NGTCP2_CRYPTO_INITIAL_SECRETLEN];
   uint8_t initial_secretbuf[NGTCP2_CRYPTO_INITIAL_SECRETLEN];
@@ -434,7 +438,6 @@ int ngtcp2_crypto_derive_and_install_initial_key(
   ngtcp2_crypto_aead_ctx retry_aead_ctx = {0};
   int rv;
   int server = ngtcp2_conn_is_server(conn);
-  uint32_t version = ngtcp2_conn_get_negotiated_version(conn);
   const uint8_t *retry_key;
   size_t retry_noncelen;
 
@@ -479,13 +482,13 @@ int ngtcp2_crypto_derive_and_install_initial_key(
   }
 
   if (ngtcp2_crypto_derive_packet_protection_key(
-          rx_key, rx_iv, rx_hp_key, &ctx.aead, &ctx.md, rx_secret,
+          rx_key, rx_iv, rx_hp_key, version, &ctx.aead, &ctx.md, rx_secret,
           NGTCP2_CRYPTO_INITIAL_SECRETLEN) != 0) {
     return -1;
   }
 
   if (ngtcp2_crypto_derive_packet_protection_key(
-          tx_key, tx_iv, tx_hp_key, &ctx.aead, &ctx.md, tx_secret,
+          tx_key, tx_iv, tx_hp_key, version, &ctx.aead, &ctx.md, tx_secret,
           NGTCP2_CRYPTO_INITIAL_SECRETLEN) != 0) {
     return -1;
   }
@@ -560,14 +563,15 @@ int ngtcp2_crypto_update_key(
   const ngtcp2_crypto_aead *aead = &ctx->aead;
   const ngtcp2_crypto_md *md = &ctx->md;
   size_t ivlen = ngtcp2_crypto_packet_protection_ivlen(aead);
+  uint32_t version = ngtcp2_conn_get_negotiated_version(conn);
 
   if (ngtcp2_crypto_update_traffic_secret(rx_secret, md, current_rx_secret,
                                           secretlen) != 0) {
     return -1;
   }
 
-  if (ngtcp2_crypto_derive_packet_protection_key(rx_key, rx_iv, NULL, aead, md,
-                                                 rx_secret, secretlen) != 0) {
+  if (ngtcp2_crypto_derive_packet_protection_key(
+          rx_key, rx_iv, NULL, version, aead, md, rx_secret, secretlen) != 0) {
     return -1;
   }
 
@@ -576,8 +580,8 @@ int ngtcp2_crypto_update_key(
     return -1;
   }
 
-  if (ngtcp2_crypto_derive_packet_protection_key(tx_key, tx_iv, NULL, aead, md,
-                                                 tx_secret, secretlen) != 0) {
+  if (ngtcp2_crypto_derive_packet_protection_key(
+          tx_key, tx_iv, NULL, version, aead, md, tx_secret, secretlen) != 0) {
     return -1;
   }
 
@@ -712,8 +716,8 @@ static int crypto_derive_token_key(uint8_t *key, size_t keylen, uint8_t *iv,
 }
 
 static size_t crypto_generate_retry_token_aad(uint8_t *dest,
-                                              const struct sockaddr *sa,
-                                              size_t salen,
+                                              const ngtcp2_sockaddr *sa,
+                                              ngtcp2_socklen salen,
                                               const ngtcp2_cid *retry_scid) {
   uint8_t *p = dest;
 
@@ -729,7 +733,7 @@ static const uint8_t retry_token_info_prefix[] = "retry_token";
 
 ngtcp2_ssize ngtcp2_crypto_generate_retry_token(
     uint8_t *token, const uint8_t *secret, size_t secretlen,
-    const struct sockaddr *remote_addr, size_t remote_addrlen,
+    const ngtcp2_sockaddr *remote_addr, ngtcp2_socklen remote_addrlen,
     const ngtcp2_cid *retry_scid, const ngtcp2_cid *odcid, ngtcp2_tstamp ts) {
   uint8_t plaintext[NGTCP2_CRYPTO_MAX_RETRY_TOKENLEN];
   uint8_t rand_data[NGTCP2_CRYPTO_TOKEN_RAND_DATALEN];
@@ -741,7 +745,7 @@ ngtcp2_ssize ngtcp2_crypto_generate_retry_token(
   ngtcp2_crypto_md md;
   ngtcp2_crypto_aead_ctx aead_ctx;
   size_t plaintextlen;
-  uint8_t aad[sizeof(struct sockaddr_storage) + NGTCP2_MAX_CIDLEN];
+  uint8_t aad[sizeof(ngtcp2_sockaddr_storage) + NGTCP2_MAX_CIDLEN];
   size_t aadlen;
   uint8_t *p = plaintext;
   int rv;
@@ -805,9 +809,9 @@ ngtcp2_ssize ngtcp2_crypto_generate_retry_token(
 
 int ngtcp2_crypto_verify_retry_token(
     ngtcp2_cid *odcid, const uint8_t *token, size_t tokenlen,
-    const uint8_t *secret, size_t secretlen, const struct sockaddr *remote_addr,
-    socklen_t remote_addrlen, const ngtcp2_cid *dcid, ngtcp2_duration timeout,
-    ngtcp2_tstamp ts) {
+    const uint8_t *secret, size_t secretlen, const ngtcp2_sockaddr *remote_addr,
+    ngtcp2_socklen remote_addrlen, const ngtcp2_cid *dcid,
+    ngtcp2_duration timeout, ngtcp2_tstamp ts) {
   uint8_t
       plaintext[/* cid len = */ 1 + NGTCP2_MAX_CIDLEN + sizeof(ngtcp2_tstamp)];
   uint8_t key[32];
@@ -817,7 +821,7 @@ int ngtcp2_crypto_verify_retry_token(
   ngtcp2_crypto_aead_ctx aead_ctx;
   ngtcp2_crypto_aead aead;
   ngtcp2_crypto_md md;
-  uint8_t aad[sizeof(struct sockaddr_storage) + NGTCP2_MAX_CIDLEN];
+  uint8_t aad[sizeof(ngtcp2_sockaddr_storage) + NGTCP2_MAX_CIDLEN];
   size_t aadlen;
   const uint8_t *rand_data;
   const uint8_t *ciphertext;
@@ -882,19 +886,19 @@ int ngtcp2_crypto_verify_retry_token(
 }
 
 static size_t crypto_generate_regular_token_aad(uint8_t *dest,
-                                                const struct sockaddr *sa) {
+                                                const ngtcp2_sockaddr *sa) {
   const uint8_t *addr;
   size_t addrlen;
 
   switch (sa->sa_family) {
   case AF_INET:
-    addr = (const uint8_t *)&((const struct sockaddr_in *)(void *)sa)->sin_addr;
-    addrlen = sizeof(((const struct sockaddr_in *)(void *)sa)->sin_addr);
+    addr = (const uint8_t *)&((const ngtcp2_sockaddr_in *)(void *)sa)->sin_addr;
+    addrlen = sizeof(((const ngtcp2_sockaddr_in *)(void *)sa)->sin_addr);
     break;
   case AF_INET6:
     addr =
-        (const uint8_t *)&((const struct sockaddr_in6 *)(void *)sa)->sin6_addr;
-    addrlen = sizeof(((const struct sockaddr_in6 *)(void *)sa)->sin6_addr);
+        (const uint8_t *)&((const ngtcp2_sockaddr_in6 *)(void *)sa)->sin6_addr;
+    addrlen = sizeof(((const ngtcp2_sockaddr_in6 *)(void *)sa)->sin6_addr);
     break;
   default:
     assert(0);
@@ -908,11 +912,10 @@ static size_t crypto_generate_regular_token_aad(uint8_t *dest,
 
 static const uint8_t regular_token_info_prefix[] = "regular_token";
 
-ngtcp2_ssize
-ngtcp2_crypto_generate_regular_token(uint8_t *token, const uint8_t *secret,
-                                     size_t secretlen,
-                                     const struct sockaddr *remote_addr,
-                                     size_t remote_addrlen, ngtcp2_tstamp ts) {
+ngtcp2_ssize ngtcp2_crypto_generate_regular_token(
+    uint8_t *token, const uint8_t *secret, size_t secretlen,
+    const ngtcp2_sockaddr *remote_addr, ngtcp2_socklen remote_addrlen,
+    ngtcp2_tstamp ts) {
   uint8_t plaintext[sizeof(ngtcp2_tstamp)];
   uint8_t rand_data[NGTCP2_CRYPTO_TOKEN_RAND_DATALEN];
   uint8_t key[32];
@@ -923,7 +926,7 @@ ngtcp2_crypto_generate_regular_token(uint8_t *token, const uint8_t *secret,
   ngtcp2_crypto_md md;
   ngtcp2_crypto_aead_ctx aead_ctx;
   size_t plaintextlen;
-  uint8_t aad[sizeof(struct sockaddr_in6)];
+  uint8_t aad[sizeof(ngtcp2_sockaddr_in6)];
   size_t aadlen;
   uint8_t *p = plaintext;
   int rv;
@@ -982,8 +985,8 @@ ngtcp2_crypto_generate_regular_token(uint8_t *token, const uint8_t *secret,
 
 int ngtcp2_crypto_verify_regular_token(const uint8_t *token, size_t tokenlen,
                                        const uint8_t *secret, size_t secretlen,
-                                       const struct sockaddr *remote_addr,
-                                       socklen_t remote_addrlen,
+                                       const ngtcp2_sockaddr *remote_addr,
+                                       ngtcp2_socklen remote_addrlen,
                                        ngtcp2_duration timeout,
                                        ngtcp2_tstamp ts) {
   uint8_t plaintext[sizeof(ngtcp2_tstamp)];
@@ -994,7 +997,7 @@ int ngtcp2_crypto_verify_regular_token(const uint8_t *token, size_t tokenlen,
   ngtcp2_crypto_aead_ctx aead_ctx;
   ngtcp2_crypto_aead aead;
   ngtcp2_crypto_md md;
-  uint8_t aad[sizeof(struct sockaddr_in6)];
+  uint8_t aad[sizeof(ngtcp2_sockaddr_in6)];
   size_t aadlen;
   const uint8_t *rand_data;
   const uint8_t *ciphertext;
@@ -1074,7 +1077,7 @@ ngtcp2_ssize ngtcp2_crypto_write_connection_close(
   }
 
   if (ngtcp2_crypto_derive_packet_protection_key(
-          tx_key, tx_iv, tx_hp_key, &ctx.aead, &ctx.md, tx_secret,
+          tx_key, tx_iv, tx_hp_key, version, &ctx.aead, &ctx.md, tx_secret,
           NGTCP2_CRYPTO_INITIAL_SECRETLEN) != 0) {
     return -1;
   }
@@ -1143,22 +1146,14 @@ ngtcp2_ssize ngtcp2_crypto_write_retry(uint8_t *dest, size_t destlen,
   return spktlen;
 }
 
-/*
- * crypto_setup_initial_crypto establishes the initial secrets and
- * encryption keys, and prepares local QUIC transport parameters.
- */
-static int crypto_setup_initial_crypto(ngtcp2_conn *conn,
-                                       const ngtcp2_cid *dcid) {
-  return ngtcp2_crypto_derive_and_install_initial_key(
-      conn, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, dcid);
-}
-
 int ngtcp2_crypto_client_initial_cb(ngtcp2_conn *conn, void *user_data) {
   const ngtcp2_cid *dcid = ngtcp2_conn_get_dcid(conn);
   void *tls = ngtcp2_conn_get_tls_native_handle(conn);
   (void)user_data;
 
-  if (crypto_setup_initial_crypto(conn, dcid) != 0) {
+  if (ngtcp2_crypto_derive_and_install_initial_key(
+          conn, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
+          ngtcp2_conn_get_original_version(conn), dcid) != 0) {
     return NGTCP2_ERR_CALLBACK_FAILURE;
   }
 
@@ -1178,9 +1173,9 @@ int ngtcp2_crypto_recv_retry_cb(ngtcp2_conn *conn, const ngtcp2_pkt_hd *hd,
                                 void *user_data) {
   (void)user_data;
 
-  if (ngtcp2_crypto_derive_and_install_initial_key(conn, NULL, NULL, NULL, NULL,
-                                                   NULL, NULL, NULL, NULL, NULL,
-                                                   &hd->scid) != 0) {
+  if (ngtcp2_crypto_derive_and_install_initial_key(
+          conn, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
+          ngtcp2_conn_get_original_version(conn), &hd->scid) != 0) {
     return NGTCP2_ERR_CALLBACK_FAILURE;
   }
 
@@ -1192,7 +1187,9 @@ int ngtcp2_crypto_recv_client_initial_cb(ngtcp2_conn *conn,
                                          void *user_data) {
   (void)user_data;
 
-  if (crypto_setup_initial_crypto(conn, dcid) != 0) {
+  if (ngtcp2_crypto_derive_and_install_initial_key(
+          conn, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
+          ngtcp2_conn_get_original_version(conn), dcid) != 0) {
     return NGTCP2_ERR_CALLBACK_FAILURE;
   }
 
