@@ -47,7 +47,6 @@
 #include "debug.h"
 #include "util.h"
 #include "shared.h"
-#include "template.h"
 
 using namespace ngtcp2;
 using namespace std::literals;
@@ -102,18 +101,9 @@ int Stream::open_file(const std::string_view &path) {
 
 namespace {
 void writecb(struct ev_loop *loop, ev_io *w, int revents) {
-  ev_io_stop(loop, w);
-
   auto c = static_cast<Client *>(w->data);
 
-  auto rv = c->on_write();
-  switch (rv) {
-  case 0:
-    return;
-  case NETWORK_ERR_SEND_BLOCKED:
-    c->start_wev();
-    return;
-  }
+  c->on_write();
 }
 } // namespace
 
@@ -125,14 +115,8 @@ void readcb(struct ev_loop *loop, ev_io *w, int revents) {
   if (c->on_read(*ep) != 0) {
     return;
   }
-  auto rv = c->on_write();
-  switch (rv) {
-  case 0:
-    return;
-  case NETWORK_ERR_SEND_BLOCKED:
-    c->start_wev();
-    return;
-  }
+
+  c->on_write();
 }
 } // namespace
 
@@ -160,25 +144,10 @@ void retransmitcb(struct ev_loop *loop, ev_timer *w, int revents) {
 
   rv = c->handle_expiry();
   if (rv != 0) {
-    goto fail;
-  }
-
-  rv = c->on_write();
-  if (rv != 0) {
-    goto fail;
-  }
-
-  return;
-
-fail:
-  switch (rv) {
-  case NETWORK_ERR_SEND_BLOCKED:
-    c->start_wev();
-    return;
-  default:
-    c->disconnect();
     return;
   }
+
+  c->on_write();
 }
 } // namespace
 
@@ -206,15 +175,7 @@ void delay_streamcb(struct ev_loop *loop, ev_timer *w, int revents) {
 
   ev_timer_stop(loop, w);
   c->on_extend_max_streams();
-
-  auto rv = c->on_write();
-  switch (rv) {
-  case 0:
-    return;
-  case NETWORK_ERR_SEND_BLOCKED:
-    c->start_wev();
-    return;
-  }
+  c->on_write();
 }
 } // namespace
 
@@ -235,7 +196,8 @@ Client::Client(struct ev_loop *loop)
       version_(0),
       early_data_(false),
       should_exit_(false),
-      handshake_confirmed_(false) {
+      handshake_confirmed_(false),
+      tx_{} {
   ev_io_init(&wev_, writecb, 0, EV_WRITE);
   wev_.data = this;
   ev_timer_init(&timer_, timeoutcb, 0.,
@@ -256,13 +218,11 @@ Client::Client(struct ev_loop *loop)
   ev_signal_init(&sigintev_, siginthandler, SIGINT);
 }
 
-Client::~Client() {
-  disconnect();
-
-  ev_io_stop(loop_, &wev_);
-}
+Client::~Client() { disconnect(); }
 
 void Client::disconnect() {
+  tx_.send_blocked = false;
+
   handle_error();
 
   config.tx_loss_prob = 0;
@@ -273,6 +233,8 @@ void Client::disconnect() {
   ev_timer_stop(loop_, &rttimer_);
   ev_timer_stop(loop_, &timer_);
 
+  ev_io_stop(loop_, &wev_);
+
   for (auto &ep : endpoints_) {
     ev_io_stop(loop_, &ep.rev);
     close(ep.fd);
@@ -589,7 +551,7 @@ int recv_new_token(ngtcp2_conn *conn, const ngtcp2_vec *token,
 
 int Client::init(int fd, const Address &local_addr, const Address &remote_addr,
                  const char *addr, const char *port, uint32_t version,
-                 const TLSClientContext &tls_ctx) {
+                 TLSClientContext &tls_ctx) {
   endpoints_.reserve(4);
 
   endpoints_.emplace_back();
@@ -754,8 +716,6 @@ int Client::init(int fd, const Address &local_addr, const Address &remote_addr,
     }
   }
 
-  ev_io_set(&wev_, fd, EV_WRITE);
-
   ev_io_start(loop_, &ep.rev);
   ev_timer_again(loop_, &timer_);
 
@@ -792,6 +752,11 @@ int Client::feed_data(const Endpoint &ep, const sockaddr *sa, socklen_t salen,
       // already set.  This is because OpenSSL might set Alert.
       last_error_ = quic_err_transport(rv);
       break;
+    case NGTCP2_ERR_CRYPTO:
+      if (!last_error_.code) {
+        process_unhandled_tls_alert();
+      }
+      // fall through
     default:
       if (!last_error_.code) {
         last_error_ = quic_err_transport(rv);
@@ -901,10 +866,17 @@ int Client::handle_expiry() {
 }
 
 int Client::on_write() {
-  if (auto rv = write_streams(); rv != 0) {
-    if (rv == NETWORK_ERR_SEND_BLOCKED) {
-      schedule_retransmit();
+  if (tx_.send_blocked) {
+    if (auto rv = send_blocked_packet(); rv != 0) {
+      return rv;
     }
+
+    if (tx_.send_blocked) {
+      return 0;
+    }
+  }
+
+  if (auto rv = write_streams(); rv != 0) {
     return rv;
   }
 
@@ -922,7 +894,6 @@ int Client::write_streams() {
   ngtcp2_vec vec;
   ngtcp2_path_storage ps;
   size_t pktcnt = 0;
-  std::array<uint8_t, 64_k> buf;
   auto max_udp_payload_size = ngtcp2_conn_get_path_max_udp_payload_size(conn_);
   size_t max_pktcnt =
       (config.cc_algo == NGTCP2_CC_ALGO_BBR ||
@@ -952,9 +923,9 @@ int Client::write_streams() {
     ngtcp2_ssize ndatalen;
     ngtcp2_pkt_info pi;
 
-    auto nwrite = ngtcp2_conn_writev_stream(conn_, &ps.path, &pi, buf.data(),
-                                            max_udp_payload_size, &ndatalen,
-                                            flags, stream_id, &vec, vcnt, ts);
+    auto nwrite = ngtcp2_conn_writev_stream(
+        conn_, &ps.path, &pi, tx_.data.data(), max_udp_payload_size, &ndatalen,
+        flags, stream_id, &vec, vcnt, ts);
     if (nwrite < 0) {
       switch (nwrite) {
       case NGTCP2_ERR_STREAM_DATA_BLOCKED:
@@ -988,28 +959,33 @@ int Client::write_streams() {
     if (nwrite == 0) {
       // We are congestion limited.
       ngtcp2_conn_update_pkt_tx_time(conn_, ts);
+      ev_io_stop(loop_, &wev_);
       return 0;
     }
 
     reset_idle_timer();
 
-    if (auto rv = send_packet(*static_cast<Endpoint *>(ps.path.user_data),
-                              ps.path.remote, pi.ecn, buf.data(), nwrite);
+    auto &ep = *static_cast<Endpoint *>(ps.path.user_data);
+
+    if (auto rv =
+            send_packet(ep, ps.path.remote, pi.ecn, tx_.data.data(), nwrite);
         rv != NETWORK_ERR_OK) {
       if (rv != NETWORK_ERR_SEND_BLOCKED) {
         last_error_ = quic_err_transport(NGTCP2_ERR_INTERNAL);
         disconnect();
-      } else {
-        ngtcp2_conn_update_pkt_tx_time(conn_, ts);
-      }
-      return rv;
-    }
 
-    ngtcp2_conn_update_pkt_tx_time(conn_, ts);
+        return rv;
+      }
+
+      ngtcp2_conn_update_pkt_tx_time(conn_, ts);
+      on_send_blocked(ep, ps.path.remote, pi.ecn, nwrite);
+
+      return 0;
+    }
 
     if (++pktcnt == max_pktcnt) {
       ngtcp2_conn_update_pkt_tx_time(conn_, ts);
-      ev_io_start(loop_, &wev_);
+      start_wev_endpoint(ep);
       return 0;
     }
   }
@@ -1395,6 +1371,65 @@ int Client::send_packet(const Endpoint &ep, const ngtcp2_addr &remote_addr,
   return NETWORK_ERR_OK;
 }
 
+void Client::on_send_blocked(const Endpoint &ep, const ngtcp2_addr &remote_addr,
+                             unsigned int ecn, size_t datalen) {
+  assert(!tx_.send_blocked);
+
+  tx_.send_blocked = true;
+
+  memcpy(&tx_.blocked.remote_addr.su, remote_addr.addr, remote_addr.addrlen);
+  tx_.blocked.remote_addr.len = remote_addr.addrlen;
+  tx_.blocked.ecn = ecn;
+  tx_.blocked.datalen = datalen;
+  tx_.blocked.endpoint = &ep;
+
+  start_wev_endpoint(ep);
+}
+
+void Client::start_wev_endpoint(const Endpoint &ep) {
+  // We do not close ep.fd, so we can expect that each Endpoint has
+  // unique fd.
+  if (ep.fd != wev_.fd) {
+    if (ev_is_active(&wev_)) {
+      ev_io_stop(loop_, &wev_);
+    }
+
+    ev_io_set(&wev_, ep.fd, EV_WRITE);
+  }
+
+  ev_io_start(loop_, &wev_);
+}
+
+int Client::send_blocked_packet() {
+  assert(tx_.send_blocked);
+
+  ngtcp2_addr remote_addr{
+      .addr = &tx_.blocked.remote_addr.su.sa,
+      .addrlen = tx_.blocked.remote_addr.len,
+  };
+
+  auto rv = send_packet(*tx_.blocked.endpoint, remote_addr, tx_.blocked.ecn,
+                        tx_.data.data(), tx_.blocked.datalen);
+  if (rv != 0) {
+    if (rv == NETWORK_ERR_SEND_BLOCKED) {
+      assert(wev_.fd == tx_.blocked.endpoint->fd);
+
+      ev_io_start(loop_, &wev_);
+
+      return 0;
+    }
+
+    last_error_ = quic_err_transport(NGTCP2_ERR_INTERNAL);
+    disconnect();
+
+    return rv;
+  }
+
+  tx_.send_blocked = false;
+
+  return 0;
+}
+
 int Client::handle_error() {
   if (!conn_ || ngtcp2_conn_is_in_closing_period(conn_)) {
     return 0;
@@ -1624,11 +1659,9 @@ int Client::select_preferred_address(Address &selected_addr,
   return 0;
 }
 
-void Client::start_wev() { ev_io_start(loop_, &wev_); }
-
 namespace {
 int run(Client &c, const char *addr, const char *port,
-        const TLSClientContext &tls_ctx) {
+        TLSClientContext &tls_ctx) {
   Address remote_addr, local_addr;
 
   auto fd = create_sock(remote_addr, addr, port);
